﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Imaging;
using System.Linq;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Mandelbrot
{
    public partial class Mandelbrot : Form
    {
        private Bitmap _image = null;

        private double _startX = -2.4;
        private double _startY = -1.3;
        private double _stopX = 1.3;
        private double _stopY = 1.3;
        private readonly Color[] _colors = new Color[100];
        private readonly List<ZoomCoordinate> _positions = new List<ZoomCoordinate>(); 

        public Mandelbrot()
        {
            InitializeComponent();
        }

        private void Mandelbrot_Load(object sender, EventArgs e)
        {
            CreateColors(new[] { Color.Black, Color.Red, Color.Yellow, Color.White },
                new [] {0, 50, 80, 100});
            RenderMandelbrot();
        }

        private void CreateColors(Color[] colors, int[] steps)
        {
            int index = 0;
            for (int i = 0; i < colors.Length - 1; i++)
            {
                int range = steps[i+1] - steps[i];
                Color start = colors[i];
                Color destination = colors[i + 1];

                for (int j = 0; j < range; j++)
                {
                    _colors[index] = Color.FromArgb((int)(start.R + (destination.R - start.R) / (double)range * j),
                       (int)(start.G + (destination.G - start.G) / (double)range * j),
                       (int)(start.B + (destination.B - start.B) / (double)range * j));
                    index++;
                }
            }
        }

        private void RenderMandelbrot()
        {
            if (pictureBox.Height == 0 || pictureBox.Width == 0)
                return;

            var startTime = DateTime.Now;
            _image = new Bitmap(pictureBox.Width, pictureBox.Height);
            
            BitmapData bData = _image.LockBits(new Rectangle(0, 0, _image.Width, _image.Height), ImageLockMode.ReadWrite, _image.PixelFormat);
            int size = bData.Stride * bData.Height;
            byte[] data = new byte[size];
            System.Runtime.InteropServices.Marshal.Copy(bData.Scan0, data, 0, size);

            var xmin = _startX;
            var ymin = _startY;
            var xmax = _stopX; 
            var ymax = _stopY; 
            double stepX = (xmax - xmin) / _image.Width; 
            double stepY = (ymax - ymin) / _image.Height;
            int height = _image.Height;

            Parallel.For(0, _image.Width, screenX =>
            {
                double x = xmin + stepX * screenX;
                double y = ymin;

                for (int screenY = 0; screenY < height; screenY++)
                {
                    double x1 = 0.0, y1 = 0.0;
                    int looper = 0;
                    while (looper < 100 && ((x1*x1) + (y1*y1)) < 4)
                    {
                        looper++;
                        double xx = (x1*x1) - (y1*y1) + x;
                        y1 = 2*x1*y1 + y;
                        x1 = xx;
                    }

                    int val = (looper == 100) ? 0 : looper;
                    var color = _colors[val];

                    data[(screenX*4 + screenY*bData.Stride) + 0] = color.B;
                    data[(screenX*4 + screenY*bData.Stride) + 1] = color.G;
                    data[(screenX*4 + screenY*bData.Stride) + 2] = color.R;
                    data[(screenX*4 + screenY*bData.Stride) + 3] = color.A;
                    y += stepY;
                }
            });

            System.Runtime.InteropServices.Marshal.Copy(data, 0, bData.Scan0, data.Length);
            _image.UnlockBits(bData);

            var timeSpan = DateTime.Now - startTime;
            Debug.WriteLine(timeSpan.TotalMilliseconds + " ms");

            pictureBox.Image = _image;
            pictureBox.Invalidate();
        }

        private void Mandelbrot_Resize(object sender, EventArgs e)
        {
            RenderMandelbrot();
        }

        private void pictureBox_MouseClick(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                // Zoom in
                double diffX = _stopX - _startX;
                double diffY = _stopY - _startY;
                double cX = _startX + (diffX) * (e.X / (double)this.Width);
                double cY = _startY + (diffY) * (e.Y / (double)this.Height);

                _positions.Add(new ZoomCoordinate(_startX, _stopX, _startY, _stopY));
                _startX = cX - diffX / 4.0;
                _stopX = cX + diffX / 4.0;
                _startY = cY - diffY / 4.0;
                _stopY = cY + diffY / 4.0;

                RenderMandelbrot();
            }
            else if (e.Button == MouseButtons.Right)
            {
                // Zoom out
                if (_positions.Count > 0)
                {
                    var pos = _positions.Last();
                    _startX = pos.StartX;
                    _stopX = pos.StopX;
                    _startY = pos.StartY;
                    _stopY = pos.StopY;
                    _positions.RemoveAt(_positions.Count - 1);
                    RenderMandelbrot();
                }
            }
        }
    }

    struct ZoomCoordinate
    {
        public double StartX;
        public double StopX;
        public double StartY;
        public double StopY;

        public ZoomCoordinate(double startX, double stopX , double startY , double stopY )
        {
            StartX = startX;
            StopX = stopX;
            StartY = startY;
            StopY = stopY;
        }
    }
}
